//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

#include <hpgcc3.h>
extern "C" {
#include <ggl.h>
}

#include "memblock.h"
#include "gstring.h"
#include "gui.h"
#include "resources.h"


// CLASS GCONTROL IMPLEMENTATION


// CREATE WITH NO PARENT, DO NOT USE DIRECTLY
gControl::gControl()
{
	// INITIALIZE TO ZERO
	memset((char *)this,0,sizeof(gControl));
	
}

gControl::gControl(gControl& _Parent)
{
	// INITIALIZE TO ZEROO
	memset((char *)this,0,sizeof(gControl));
	
	// ESTABLISH LINK TO PARENT
	_Parent.AddChild(*this);		
	
}

gControl::~gControl()
{
	if(Parent!=NULL) Parent->RemoveChild(*this);
	
	// FREE ALLOCATED RESOURCES
	
	if(Events)
	{
		gEvHandlerList *ptr;
		while(Events) { ptr=Events; Events=Events->Next; free(ptr); }
	}
	
	if(drawsurf.addr) { free(drawsurf.addr); drawsurf.addr=NULL; }	// MARK NULL FOR OTHER DESTRUCTORS
	
}

void gControl::AddChild(gControl& child)
{
	child.Parent=this;
	if(FirstChild==NULL) FirstChild=LastChild=&child;
	else {
		LastChild->Next=&child;
		child.Prev=LastChild;
		LastChild=&child;
	}
}

void gControl::RemoveChild(gControl& child)
{
gControl *ptr=FirstChild;

while(ptr && (ptr!=&child)) ptr=ptr->Next;
if(!ptr) return;

if(ptr->Next) ptr->Next->Prev=ptr->Prev;
if(ptr->Prev) ptr->Prev->Next=ptr->Next;
if(FirstChild==ptr) FirstChild=ptr->Next;
if(LastChild==ptr) LastChild=ptr->Prev;

ptr->Parent=NULL;
ptr->Next=ptr->Prev=NULL;
}


BOOL gControl::DefaultEvent(gEvent *e)
{
	// PERFORM ALL BASIC EVENTS
	
return FALSE;	
}


BOOL gControl::DoEvent(gEvent *e)
{
	
	if(DoEventExclusive(e)) return TRUE;
	
gControl *ptr=FirstChild;

while(ptr) {
	if(ptr->DoEvent(e)) return TRUE;
	ptr=ptr->Next;
	}
return FALSE;	
}

BOOL gControl::DoEventExclusive(gEvent *e)
{
gEvHandlerList *ptr=Events;
while(ptr) {
	if( (e->Type==ptr->Type) && ((e->Arg&ptr->Mask)==ptr->Value)) return (ptr->Action)(e,this);
	ptr=ptr->Next;
}
return DefaultEvent(e);
}


void gControl::AddHandler(unsigned int Type,unsigned int Mask,unsigned int Arg,gEvHandler Function)
{
	gEvHandlerList *ap=new gEvHandlerList;
	
	if(!ap) InsufMemoryError();
	ap->Type=Type;
	ap->Mask=Mask;
	ap->Value=Arg;
	ap->Action=Function;
	
	ap->Next=Events;
	Events=ap;
	
}
void gControl::RemoveHandler(unsigned int Type,unsigned int Mask,unsigned int Arg)
{
	gEvHandlerList *prev,*ptr;
	
	prev=ptr=Events;
	
	while(ptr) {
		if( (ptr->Type==Type) && (ptr->Mask==Mask) && (ptr->Value==Arg)) {
			prev->Next=ptr->Next;
			if(Events==ptr) Events=ptr->Next;
			delete ptr;
			if(prev==ptr) { ptr=Events; continue; }
			else ptr=prev;
		}
		prev=ptr;
		ptr=ptr->Next;
	}
	
}
gEvHandler gControl::GetHandler(unsigned int Type,unsigned int Mask,unsigned int Arg)
{
	gEvHandlerList *ptr;
	
	ptr=Events;
	
	while(ptr) {
		if(ptr->Type==Type && ptr->Mask==Mask && ptr->Value==Arg) {
			return ptr->Action;
		}
		ptr=ptr->Next;
	}

	return NULL;
}

















void gControl::InsufMemoryError()
{
	sys_error("Insufficient memory",-1);
}


// GRAPHIC MANAGEMENT
BOOL gControl::Resize(int width,int height)
{
	int size=(width*height+1)>>1;
	
	char *buffer=(char *)malloc(size);
	if(!buffer) return FALSE;
	
	memset(buffer,0,size);
	
	if(drawsurf.addr) {
		gglsurface dest;
		dest.addr=(int *)buffer;
		dest.width=width;
		dest.x=0;
		dest.y=0;
		drawsurf.x=0;
		drawsurf.y=0;
		int w,h;
		w=min(width,sizex);
		h=min(height,sizey);
		ggl_bitblt(&dest,&drawsurf,w,h);
		free(drawsurf.addr);
	}
	drawsurf.addr=(int *)buffer;
	drawsurf.width=width;
	gUpdate u;
	u.clipx=0;
	u.clipy=0;
	u.clipx2=max(sizex,width)-1;
	u.clipy2=max(sizey,height)-1;
	sizex=width;
	sizey=height;
	Invalidate(&u);

	return TRUE;
}

void gControl::Move(int x,int y)
{
	int x2=posx;
	int y2=posy;
	posx=x;
	posy=y;
	gUpdate u;
	u.clipx=x2;
	u.clipy=y2;
	u.clipx2=x2+sizex-1;
	u.clipy2=y2+sizey-1;
	
	Invalidate(&u);
	u.clipx=posx;
	u.clipy=posy;
	u.clipx2=posx+sizex-1;
	u.clipy2=posy+sizey-1;
	Invalidate(&u);
	
}

void gControl::ParentToClient(int &x,int &y)
{
	x-=posx-viewx;
	x-=posy-viewy;
}
void gControl::ClientToParent(int &x,int &y)
{
	x+=posx-viewx;
	y+=posy-viewy;
}


// FLAT COLOR ROUTINES
void gControl::HLine(int y,int x1,int x2,int color)
{
	// CLIP LINE
	if(clipx<0) return;
	
	if(y<clipy) return;
	if(y>clipy2) return;
	if(x1>x2) {
		int temp=x1;
		x1=x2;
		x2=temp;
	}
	if(x1>clipx2) return;
	if(x2<clipx) return;
	if(x1<clipx) x1=clipx;
	if(x2>clipx2) x2=clipx2;
	ggl_hline(&drawsurf,y,x1,x2,color);
}
void gControl::VLine(int x,int y1,int y2,int color)
{
	if(clipx<0) return;

	if(x<clipx) return;
	if(x>clipx2) return;
	
	if(y1>y2) {
		int temp=y1;
		y1=y2;
		y2=temp;
	}
	
	if(y1>clipy2) return;
	if(y2<clipy) return;
	if(y1<clipy) y1=clipy;
	if(y2>clipy2) y2=clipy2;
	
	ggl_vline(&drawsurf,x,y1,y2,color);
	
}

void gControl::Line(int x,int y,int x2,int y2,int color)
{
	// TODO: LINE W/BRESSENHAM AND CLIPPING
	if(clipx<0) return;
	
}
void gControl::Rect(int x1,int y1,int x2,int y2,int color)
{
	if(clipx<0) return;

	if(y1>y2) {
		int temp=y1;
		y1=y2;
		y2=temp;
	}
	if(x1>x2) {
		int temp=x1;
		x1=x2;
		x2=temp;
	}
	if(x2<clipx) return;
	if(x1>clipx2) return;
	if(y1>clipy2) return;
	if(y2<clipy) return;
	if(x2>clipx2) x2=clipx2;
	if(x1<clipx) x1=clipx;
	if(y1<clipy) y1=clipy;
	if(y2>clipy2) y2=clipy2;
	ggl_rect(&drawsurf,x1,y1,x2,y2,color);
	
}

void gControl::Plot(int x1,int y1,int color)
{
	if(clipx<0) return;


	if(x1>clipx2) return;
	if(y1>clipy2) return;
	if(x1<clipx) return;
	if(y1<clipy) return;

	ggl_pltnib(drawsurf.addr,x1+drawsurf.width*y1,color);
	
}

int gControl::GetPixel(int x1,int y1)
{
	if(x1>=sizex) return 0;
	if(y1>=sizey) return 0;
	if(x1<0) return 0;
	if(y1<0) return 0;

	return ggl_getnib(drawsurf.addr,x1+drawsurf.width*y1);
	
}


void gControl::RectP(int x1,int y1,int x2,int y2,int *Pattern)
{
	if(clipx<0) return;

	if(y1>y2) {
		int temp=y1;
		y1=y2;
		y2=temp;
	}
	if(x1>x2) {
		int temp=x1;
		x1=x2;
		x2=temp;
	}
	if(x2<clipx) return;
	if(x1>clipx2) return;
	if(y1>clipy2) return;
	if(y2<clipy) return;
	if(x2>clipx2) x2=clipx2;
	if(x1<clipx) x1=clipx;
	if(y1<clipy) y1=clipy;
	if(y2>clipy2) y2=clipy2;
	ggl_rectp(&drawsurf,x1,y1,x2,y2,Pattern);
}


unsigned int gui_chgcolorfilter(unsigned int dest,unsigned int src,int param)
{
	if(!param) return ggl_opmask(dest,~src,0xffffffff);
	return ggl_opmask(dest,src&param,0);
}

// TEXT OUTPUT ROUTINES
void gControl::DrawText(int x,int y,char *Text,gFont *Font,int color)
{

	// CLIPPING
	if(clipx<0) return;
	
	if(y>clipy2) return;
	if(y+(int)Font->FontData->BitmapHeight<=clipy) return;

	int w,h;
	gglsurface srf;
	srf.addr=(int *)Font->FontData->Bitmap;
	srf.width=Font->FontData->BitmapWidth;
	srf.y=0;

	h=Font->FontData->BitmapHeight;
	
	if(y<clipy) {
		h-=clipy-y;
		srf.y=clipy-y;
		y=clipy;
	}
	if(y+h-1>clipy2) h=clipy2-y+1;

	drawsurf.y=y;
	drawsurf.x=x;
	
	while(*Text) {
		if(*Text=='\n' || *Text=='\r') return;
	w=Font->FontData->WidthMap[(int)(*Text)];
	srf.x=w&0xffff;
	w>>=16;
	if(drawsurf.x>clipx2) return;
	if(drawsurf.x+w-1<clipx) return;
	if(drawsurf.x<clipx) {
			srf.x+=clipx-drawsurf.x;
			w-=clipx-drawsurf.x;
			drawsurf.x=clipx;
	}
	if(drawsurf.x+w-1>clipx2) w=clipx2-drawsurf.x+1;
	
	if((color&0xf)==0xf) ggl_bitbltmask(&drawsurf,&srf,w,h,0);
	else ggl_bitbltoper(&drawsurf,&srf,w,h,ggl_mkcolor(color&0xf),&gui_chgcolorfilter);
	drawsurf.x+=w;
	++Text;	

	}
	return;
	

}

void gControl::DrawTextBk(int x,int y,char *Text,gFont *Font,int color,int bkcolor)
{

	// CLIPPING
	if(clipx<0) return;
	
	if(y>clipy2) return;
	if(y+(int)Font->FontData->BitmapHeight<=clipy) return;

	int w,h;
	gglsurface srf;
	srf.addr=(int *)Font->FontData->Bitmap;
	srf.width=Font->FontData->BitmapWidth;
	srf.y=0;

	h=Font->FontData->BitmapHeight;
	
	if(y<clipy) {
		h-=clipy-y;
		srf.y=clipy-y;
		y=clipy;
	}
	if(y+h-1>clipy2) h=clipy2-y+1;

	drawsurf.y=y;
	drawsurf.x=x;
	
	while(*Text) {
		if(*Text=='\n' || *Text=='\r') return;
	w=Font->FontData->WidthMap[(int)(*Text)];
	srf.x=w&0xffff;
	w>>=16;
	if(drawsurf.x>clipx2) return;
	if(drawsurf.x+w-1<clipx) return;
	if(drawsurf.x<clipx) {
			srf.x+=clipx-drawsurf.x;
			w-=clipx-drawsurf.x;
			drawsurf.x=clipx;
	}
	if(drawsurf.x+w-1>clipx2) w=clipx2-drawsurf.x+1;
	
	ggl_rect(&drawsurf,drawsurf.x,drawsurf.y,drawsurf.x+w-1,drawsurf.y+h-1,ggl_mkcolor(bkcolor&0xf));
	if((color&0xf)==0xf) ggl_bitbltmask(&drawsurf,&srf,w,h,0);
	else ggl_bitbltoper(&drawsurf,&srf,w,h,ggl_mkcolor(color&0xf),&gui_chgcolorfilter);
	drawsurf.x+=w;
	++Text;	

	}
	return;
	

}




// BITMAPS AND ICON ROUTINES
void gControl::DrawIcon(int x,int y,unsigned int *IconData,int width,int height)

{

	if(clipx<0) return;

	gglsurface src;
	
	src.addr=(int *)IconData;
	src.width=width;
	src.x=0;
	src.y=0;
	
	// CLIPPING
	
	if(x+width-1<clipx) return;
	if(x>clipx2) return;
	if(y>clipy2) return;
	if(y+height-1<clipy) return;

	
	if(x<clipx) {
		src.x=clipx-x;
		x=clipx;
		width-=src.x;
	}
	if(x+width-1>clipx2) {
		width=clipx2-x+1;
	}

	if(y<clipy) {
		src.y=clipy-y;
		y=clipy;
		height-=src.y;
	}
	if(y+height-1>clipy2) {
		height=clipy2-y+1;
	}
	
	drawsurf.x=x;
	drawsurf.y=y;
	ggl_bitblt(&drawsurf,&src,width,height);
	
	
}

// BITMAPS AND ICON ROUTINES
void gControl::DrawIconPartial(int x,int y,unsigned int *IconData,int iconwidth,int iconx,int icony,int width,int height)

{

	if(clipx<0) return;

	gglsurface src;
	
	src.addr=(int *)IconData;
	src.width=iconwidth;
	src.x=iconx;
	src.y=icony;
	
	// CLIPPING
	
	if(x+width-1<clipx) return;
	if(x>clipx2) return;
	if(y>clipy2) return;
	if(y+height-1<clipy) return;

	
	if(x<clipx) {
		src.x=clipx-x;
		x=clipx;
		width-=src.x;
	}
	if(x+width-1>clipx2) {
		width=clipx2-x+1;
	}

	if(y<clipy) {
		src.y=clipy-y;
		y=clipy;
		height-=src.y;
	}
	if(y+height-1>clipy2) {
		height=clipy2-y+1;
	}
	
	drawsurf.x=x;
	drawsurf.y=y;
	ggl_bitblt(&drawsurf,&src,width,height);
	
	
}



// CLIPPING ROUTINES
void gControl::SetClipRegion(int x1,int y1,int x2,int y2)
{
	if(x1>x2) {
		int tmp=x1;
		x1=x2;
		x2=tmp;
	}
	if(y1>y2) {
		int tmp=y1;
		y1=y2;
		y2=tmp;
	}
	
	if(x1<0) x1=0;
	if(x1>=sizex) { clipx=-1; return; }

	if(x2<0) { clipx=-1; return; }
	if(x2>=sizex) x2=sizex-1;

	if(y1<0) y1=0;
	if(y1>=sizey) { clipx=-1; return; }
	
	if(y2<0) { clipx=-1; return; }
	if(y2>=sizey) y2=sizey-1;
	
	clipx=x1;
	clipy=y1;
	clipx2=x2;
	clipy2=y2;
	
}

void gControl::ClearClipRegion()
{
	clipx=0;
	clipy=0;
	clipx2=sizex-1;
	clipy2=sizey-1;
}

void gControl::ClipToView(gUpdate *u)
{
	if(u->clipx<viewx) u->clipx=viewx;
	if(u->clipx>viewx+vieww-1) { u->clipx=-1; return; }
	if(u->clipx2<viewx) { u->clipx=-1; return; } 
	if(u->clipx2>viewx+vieww-1) u->clipx2=viewx+vieww-1;

	if(u->clipy<viewy) u->clipy=viewy;
	if(u->clipy>viewy+viewh-1) { u->clipx=-1; return; }
	if(u->clipy2<viewy) { u->clipx=-1; return; } 
	if(u->clipy2>viewy+viewh-1) u->clipy2=viewy+viewh-1;

}



void gControl::Invalidate(gControl *Sender)
{
	if(Sender) {
	if(UpdateFirst) {
		gControl *ch=Sender;
		while(ch) { if(UpdateFirst==ch) { UpdateFirst=Sender; break; } ch=ch->Next; }
	}
	else UpdateFirst=Sender;
	} else UpdateFirst=NULL;
	
	gUpdate u;
	
	if(vieww!=0) {
		u.clipx=viewx;
		u.clipy=viewy;
		u.clipx2=viewx+vieww-1;
		u.clipy2=viewy+viewh-1;
	}
	else {
	u.clipx=u.clipy=0;
	u.clipx2=sizex-1;
	u.clipy2=sizey-1;
	}
	ClientToParent(u.clipx,u.clipy);
	ClientToParent(u.clipx2,u.clipy2);
	
	Parent->Invalidate(&u,this);
	
}

void gControl::Invalidate(gUpdate *_u,gControl *Sender)
{
	if(Sender) {
	if(UpdateFirst) {
		gControl *ch=Sender;
		while(ch) { if(UpdateFirst==ch) { UpdateFirst=Sender; break; } ch=ch->Next; }
	}
	else UpdateFirst=Sender;
	} else UpdateFirst=NULL;
	
	ClipToView(_u);
	if(_u->clipx==-1) return;
	ClientToParent(_u->clipx,_u->clipy);
	ClientToParent(_u->clipx2,_u->clipy2);
	
	Parent->Invalidate(_u,this);
	
}


void gControl::RedrawBackground(gUpdate *u)
{

	if(RedrawBk!=NULL) {
	
		gEvent e;
			e.Type=GEVENT_GUI;
			e.Arg=0;
			e.Data=(unsigned int)u;
			
			(*RedrawBk)(&e,this);

	}

}

void gControl::Update(gUpdate *u)
{

	gUpdate uclient;
	
	uclient.clipx=u->clipx-posx+viewx;
	uclient.clipy=u->clipy-posy+viewy;
	uclient.clipx2=u->clipx2-posx+viewx;
	uclient.clipy2=u->clipy2-posy+viewy;

	// PASS UPDATE MESSAGE TO CHILDS
	SetClipRegion(uclient.clipx,uclient.clipy,uclient.clipx2,uclient.clipy2);
	
	gControl *chld=(UpdateFirst? UpdateFirst:FirstChild);
	
	while(chld) {
		chld->Update(&uclient);
		chld=chld->Next;
	}
	
	ClearClipRegion();
	
	// UPDATE VIEWED PORTION
	
	if(Parent) {
	drawsurf.x=(u->clipx<posx)? viewx:(u->clipx-posx+viewx);
	drawsurf.y=(u->clipy<posy)? viewy:(u->clipy-posy+viewy);
	
	if(drawsurf.x>=viewx+vieww) return;
	if(drawsurf.y>=viewy+viewh) return;
	
	int dw,dh;
	dw=u->clipx2-posx+viewx-drawsurf.x;
	if(dw<=0) return;
	if(dw>vieww) dw=vieww;
	
	dh=u->clipy2-posy+viewy-drawsurf.y;
	if(dh<=0) return;
	if(dh>viewh) dh=viewh;
	
	
	Parent->drawsurf.x=drawsurf.x-viewx+posx;
	Parent->drawsurf.y=drawsurf.y-viewy+posy;
	
	ggl_bitblt(&(Parent->drawsurf),&drawsurf,dw,dh);
	}

	UpdateFirst=NULL;
}

void gControl::SetViewY(int cy)
{
	if(cy+viewh>sizey) cy=sizey-viewh;
	if(cy<0) cy=0;
	
	if(viewy!=cy) {
		Invalidate();
	if(scrollv) {
		int x,y;
		gUpdate u;

		x=posx+vieww;
		y=posy;
		u.clipx=x;
		u.clipy=y;
		u.clipx2=x+SCROLLBARWIDTH-1;
		u.clipy2=y+viewh-1;
		Parent->Invalidate(&u,this);
	}
}
	viewy=cy;
	return;
}

void gControl::SetViewX(int cy)
{
	if(cy+vieww>sizex) cy=sizex-vieww;
	if(cy<0) cy=0;
	
	if(viewx!=cy) {
		Invalidate();
		if(scrollh) {
			int x,y;
			gUpdate u;

			x=posx;
			y=posy+viewh;
			u.clipx=x;
			u.clipy=y;
			u.clipx2=x+vieww-1;
			u.clipy2=y+SCROLLBARWIDTH-1;
			Parent->Invalidate(&u,this);
		}
	}
	viewx=cy;
	return;
}



void gControl::DrawVScrollBar()
{
	gControl *P=Parent;

	int x,y,clen,cpos;

	x=posx+vieww;
	y=posy;
	
	// DRAW FRAME
	P->VLine(x,y,y+viewh-1,GGRAY7);
	P->Rect(x+1,y,x+SCROLLBARWIDTH-2,y+viewh-1,GGRAY3);
	P->HLine(y,x+1,x+SCROLLBARWIDTH-1,GGRAY7);
	P->VLine(x+SCROLLBARWIDTH-1,y,y+viewh-1,GWHITE);
	P->HLine(y+viewh-1,x+1,x+SCROLLBARWIDTH-1,GWHITE);

	// DRAW CURSOR
	if(sizey>0) clen=((viewh-1)*(viewh-1))/sizey;
	else return;
	if(clen<1) clen=1;
	cpos=(viewy*viewh)/sizey;
	
	P->Rect(x+1,y+cpos+1,x+SCROLLBARWIDTH-2,y+cpos+clen+1,GBLACK);

}
		


void gControl::DrawHScrollBar()
{
	gControl *P=Parent;

	int x,y,clen,cpos;

	x=posx;
	y=posy+viewh;
	
	// DRAW FRAME
	P->HLine(y,x,x+vieww-1,GGRAY7);
	P->Rect(x,y+1,x+vieww-1,y+SCROLLBARWIDTH-2,GGRAY3);
	P->VLine(x,y+1,y+SCROLLBARWIDTH-1,GGRAY7);
	P->HLine(y+SCROLLBARWIDTH-1,x,x+vieww-1,GWHITE);
	P->VLine(x+vieww-1,y+1,y+SCROLLBARWIDTH-1,GWHITE);

	// DRAW CURSOR
	if(sizex>0) clen=((vieww-2)*(vieww-2))/sizex;
	else return;
	if(clen<1) clen=1;
	cpos=(viewx*vieww)/sizex;
	
	P->Rect(x+cpos+1,y+1,x+cpos+clen+1,y+SCROLLBARWIDTH-2,GBLACK);



}


gApplication *gControl::GetApplication()
{
	gControl *ptr=Parent;
	while(ptr) {
		if(ptr->Parent==NULL) return (gApplication *)ptr;
		ptr=ptr->Parent;
	}
	
	return NULL;
	
}

