//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***



// $Id: sat3_carray.c 547 2008-02-06 01:01:02Z claudio $

#include <hpgcc3.h>
#include <master_globals.h>

extern ULONGLONG __sat3_dbl2real(double value);
extern ULONGLONG __sat3_xreal2real(XREAL *value);
extern ULONGLONG __sat3_decn2real(decNumber *value);

extern double __sat3_real2dbl(SAT_OBJECT realbody);
extern void __sat3_real2xreal(SAT_OBJECT realbody,XREAL *result);
extern void __sat3_real2decn(SAT_OBJECT realbody,decNumber *result);

extern SAT_OBJECT __sat3_arrayget(SAT_OBJECT array,int row,int col);

BOOL sat3_iscarray(SAT_OBJECT array)
{
	if(!array) return 0;
	if(sat3_isptr(array)) array=sat3_peek(array,5);
	if(sat3_peek(array,5)!=SAT_DOARRY) return 0;		// RETURN IF NOT AN ARRAY
	if(sat3_peek(array+10,5)==SAT_DOCMP) return 1;
	return 0;
}



SAT_OBJECT sat3_createcarray(SAT_OBJECT where,int rows,int cols)
{
int dims=( (rows && cols)? 2: (rows || cols)? 1:0);
int size;
SAT_OBJECT ptr;

	if(!dims) return 0;
	if(dims==1) { rows+=cols; cols=1; }		// SUPPORT FOR ONE-DIMENSIONAL VECTORS BY MAKING EITHER ROWS OR COLS =0
	
	size=20+5*dims+32*rows*cols;
    where=sat3_alloc(where,size);
	if(where==0) return 0;
	
	sat3_poke(where,SAT_DOARRY,5);
    sat3_poke(where+5,size-5,5);
	sat3_poke(where+10,SAT_DOCMP,5);
	ptr=where+15;
	sat3_poke(ptr,dims,5);
	sat3_poke(ptr+5,rows,5);
	ptr+=10;
	if(dims==2) { sat3_poke(ptr,cols,5); ptr+=5; }
	sat3_memset(ptr,0,8,where+size-ptr);

    return where;
}






void sat3_carrayput_dbl(SAT_OBJECT array,int row,int col,double rpart,double ipart)
{
ULONGLONG real,imag;

	if(!sat3_iscarray(array)) return;
	
	array=__sat3_arrayget(array,row,col);
	
	if(!array) return;
	
	real=__sat3_dbl2real(rpart);
	imag=__sat3_dbl2real(ipart);
    
    sat3_poke(array,(unsigned int)real,8);
    sat3_poke(array+8,(unsigned int)(real>>32),8);
    sat3_poke(array+16,(unsigned int)imag,8);
    sat3_poke(array+24,(unsigned int)(imag>>32),8);

}



void sat3_carrayput_xreal(SAT_OBJECT array,int row,int col,XREAL *rpart,XREAL *ipart)
{
ULONGLONG real,imag;

	if(!sat3_iscarray(array)) return;
	
	array=__sat3_arrayget(array,row,col);
	
	if(!array) return;
	
	real=__sat3_xreal2real(rpart);
	imag=__sat3_xreal2real(ipart);
    
    sat3_poke(array,(unsigned int)real,8);
    sat3_poke(array+8,(unsigned int)(real>>32),8);
    sat3_poke(array+16,(unsigned int)imag,8);
    sat3_poke(array+24,(unsigned int)(imag>>32),8);

}



void sat3_carrayput_int(SAT_OBJECT array,int row,int col,int rpart,int ipart)
{
XREAL x,i;

xr_int2xr(rpart,&x);
xr_int2xr(ipart,&i);

sat3_carrayput_xreal(array,row,col,&x,&i);


}

void sat3_carrayput_llong(SAT_OBJECT array,int row,int col,LONGLONG rpart,LONGLONG ipart)
{
XREAL x,i;

xr_llong2xr(rpart,&x);
xr_llong2xr(ipart,&i);

sat3_carrayput_xreal(array,row,col,&x,&i);


}

void sat3_carrayput_decn(SAT_OBJECT array,int row,int col,decNumber *rpart,decNumber *ipart)
{
ULONGLONG real,imag;

	if(!sat3_iscarray(array)) return;
	
	array=__sat3_arrayget(array,row,col);
	
	if(!array) return;
	
	real=__sat3_decn2real(rpart);
	imag=__sat3_decn2real(ipart);
    
    sat3_poke(array,(unsigned int)real,8);
    sat3_poke(array+8,(unsigned int)(real>>32),8);
    sat3_poke(array+16,(unsigned int)imag,8);
    sat3_poke(array+24,(unsigned int)(imag>>32),8);

}

#define MAX_IEEE754_EXPO 307





BOOL sat3_carrayget_dbl(SAT_OBJECT array,int row,int col,double *rpart,double *ipart)
{

	if(!sat3_iscarray(array)) return FALSE;
	
	array=__sat3_arrayget(array,row,col);
	
	if(!array) return FALSE;
	
	*rpart=__sat3_real2dbl(array);
	*ipart=__sat3_real2dbl(array+16);
	return TRUE;

}


BOOL sat3_carrayget_xreal(SAT_OBJECT array,int row,int col,XREAL *rpart,XREAL *ipart)
{

	if(!sat3_iscarray(array)) return FALSE;
	
	array=__sat3_arrayget(array,row,col);
	
	if(!array) return FALSE;
	
	__sat3_real2xreal(array,rpart);
	__sat3_real2xreal(array+16,ipart);
	return TRUE;

}

BOOL sat3_carrayget_int(SAT_OBJECT array,int row,int col,int *rpart,int *ipart)
{
XREAL x,i;
	if(!sat3_carrayget_xreal(array,row,col,&x,&i)) return FALSE;
	
	*rpart=xr_xr2int(&x);
	*ipart=xr_xr2int(&i);
	return TRUE;
}

BOOL sat3_carrayget_llong(SAT_OBJECT array,int row,int col,LONGLONG *rpart,LONGLONG *ipart)
{
XREAL x,i;
	if(!sat3_carrayget_xreal(array,row,col,&x,&i)) return FALSE;
	
	*rpart=xr_xr2llong(&x);
	*ipart=xr_xr2llong(&i);
	return TRUE;
}


BOOL sat3_carrayget_decn(SAT_OBJECT array,int row,int col,decNumber *rpart,decNumber *ipart)
{

	if(!sat3_iscarray(array)) return FALSE;
	
	array=__sat3_arrayget(array,row,col);
	
	if(!array) return FALSE;
	
	__sat3_real2decn(array,rpart);
	__sat3_real2decn(array+16,ipart);
	return TRUE;

}

BOOL sat3_carraysize(SAT_OBJECT array,int *row,int *col)
{
	if(!sat3_iscarray(array)) return 0;
	int rows,cols,dims;

	dims=sat3_peek(array+15,5);
	if(dims<1 || dims>2) return 0;					// ONLY 1 OR 2 DIMS SUPPORTED
	array+=20;
	rows=sat3_peek(array,5);
	if(dims>=2) { array+=5; cols=sat3_peek(array,5); }
	else cols=0;
	*row=rows;
	*col=cols;
	
	return TRUE;
}
