//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***



// $Id: sat3_dir.c 486 2007-10-21 21:17:04Z claudio $

#include <hpgcc3.h>
#include <master_globals.h>


extern int __sat3_getblockinfo(SAT_OBJECT ptr,SAT_OBJECT *blkstart,int *blksize);

SAT_OBJECT __sat3_dirskiplibs(SAT_OBJECT objaddr)
{
	
    SAT_OBJECT home=sat3_peek(SAT_USEROB,5)+5;
    
    objaddr+=5;
    
    // SKIP LIBRARIES ATTACHED TO HOME
    if(objaddr==home) objaddr+=13*sat3_peek(objaddr,3);
    
    return objaddr+3;

}

// GET HOME DIR

SAT_OBJECT sat3_home()
{
    return sat3_peek(SAT_USEROB,5);
}

// GET CURRENT WORK DIR

SAT_OBJECT sat3_getcwd()
{
    return sat3_peek(SAT_CONTEXT,5);
}

// SET CURRENT WORK DIR

BOOL sat3_setcwd(SAT_OBJECT dir)
{
	if(!dir) return 0;
	if((dir&0x60000000)!=0) return 0;	// CANNOT CHANGE TO A DIR IN C-HEAP
	if(sat3_peek(dir,5)!=SAT_DORRP) return 0;
    sat3_poke(SAT_CONTEXT,dir,5);
    return 1;
}




// GET FIRST OBJECT IN A DIR

SAT_OBJECT sat3_findfirst(SAT_OBJECT dir)
{
	if(!dir) return 0;
	if(sat3_peek(dir,5)!=SAT_DORRP) return 0;
	if( !(dir=__sat3_dirskiplibs(dir)) ) return 0;
    
    int offset=sat3_peek(dir,5);
    
    if(!offset) return 0;
    
    return sat3_sknameup(dir+offset);
}

// GET NEXT OBJECT IN A DIR, 0 IF NOT MORE OBJECTS

SAT_OBJECT sat3_findnext(SAT_OBJECT dir, SAT_OBJECT diritem)
{
	SAT_OBJECT dirfirst=sat3_findfirst(dir);
	if(!dirfirst) return 0;
	if(diritem>dirfirst || diritem<dir) return 0;	// CHECK IF GIVEN ITEM BELONGS TO A DIR
    int offset;
    diritem=sat3_sknamedn(diritem)-5;

	if(diritem>dirfirst || diritem<dir) return 0;    // CHECK IF STILL WITHIN DIR

    offset=sat3_peek(diritem,5);

    if(offset==0) return 0;
    
    diritem-=offset;
    
 	if(diritem>dirfirst || diritem<dir) return 0;    // CHECK AGAIN IF WITHIN DIRECTORY   
    
    return sat3_sknameup(diritem);

}

// GET DIRECTORY THAT CONTAINS A GIVEN OBJECT
SAT_OBJECT __sat3_getdir(SAT_OBJECT var)
{
SAT_OBJECT dir;
SAT_OBJECT obj,end;
if(!var) return 0;
if((var&0x60000000)!=0) {
SAT_OBJECT 	blk;
int blklen;
	if(__sat3_getblockinfo(var,&blk,&blklen)==INVALID_PTR) return 0;
	if(sat3_peek(blk,5)!=SAT_DORRP) return 0;
	dir=blk;
}
else dir=sat3_home();

if(var==dir) return 0;		// RETURN 0 WHEN TRYING TO GET THE PARENT OF HOME

while(dir) {
	obj=sat3_findfirst(dir);
	do {
	end=sat3_skipob(obj);
	if(var==obj) return dir;
	if(var>obj && var<end) {	// OBJECT IS WITHIN DIR
		if(sat3_peek(obj,5)==SAT_DORRP) break;
		else return obj;		// VAR IS NOT A VARIABLE, BUT IS EMBEDDED IN A VAR
	}
	obj=sat3_findnext(dir,obj);
	} while(obj);
	dir=obj;
}
	
return 0;	
	
}
// GET VARIABLE NAME GIVEN THE OBJECT IF IN USEROB

BOOL sat3_getvarname(SAT_OBJECT obj,char *namebuffer, int max_chars)
{
SAT_OBJECT dir,item;

dir=__sat3_getdir(obj);

if(!dir) return 0;

item=sat3_findfirst(dir);

while(item)
{
	if(obj>item) return 0;	// ITEM MAYBE EMBEDDED IN A VAR, BUT IT'S NOT A VAR ITSELF
	if(obj==item) {
	item=sat3_sknamedn(item);
	
	int nchars=sat3_peek(item,2);
	if(max_chars<nchars+1) return 0;		// NOT ENOUGH SPACE FOR THE STRING
	item+=2;
	while(nchars--) { *namebuffer++=sat3_peek(item,2); item+=2; }
	*namebuffer=0;
	return 1; 
		
	}
	item=sat3_findnext(dir,item);	
}

return 0;		// NOT FOUND IN DIR
	
}


int sat3_getvarnamelen(SAT_OBJECT obj)
{
SAT_OBJECT dir,item;

dir=__sat3_getdir(obj);

if(!dir) return -1;

item=sat3_findfirst(dir);

while(item)
{
	if(obj>item) return -1;	// ITEM MAYBE EMBEDDED IN A VAR, BUT IT'S NOT A VAR ITSELF
	if(obj==item) {
	item=sat3_sknamedn(item);
	
	return sat3_peek(item,2);
		
	}
	item=sat3_findnext(dir,item);	
}

return -1;		// NOT FOUND IN DIR
	
}




SAT_OBJECT sat3_updir(SAT_OBJECT dir)
{
return __sat3_getdir(dir);	
}







SAT_OBJECT __sat3_rclindir(SAT_OBJECT dir,char *name,int len)
{
	int nchars;
	char *tmp;
	
	
	SAT_OBJECT obj,objname;
	
	obj=sat3_findfirst(dir);
	
	while(obj) {
	objname=sat3_sknamedn(obj);
	
	nchars=sat3_peek(objname,2);
	if(len==nchars) {	// IF LENGTH IS THE SAME, COMPARE STRINGS
		
	objname+=2;
	tmp=name;
	while(nchars--) { if(*tmp!=sat3_peek(objname,2)) break; objname+=2; ++tmp;}
	
	if(nchars<0) { return obj; }	// FOUND MATCHING NAME
	
	}
	obj=sat3_findnext(dir,obj);	
		
	}

	return 0;
	
}


// SAME AS RCL, BUT DON'T RETURN VARS IN PARENT DIRS
SAT_OBJECT sat3_rclindir(SAT_OBJECT dir,char *name)
{
	
	if(!name) return 0;
	
	char *start,*tmp;
	SAT_OBJECT curdir=dir,obj;
	tmp=name;
	start=tmp;
	
	while(*tmp!=0) {
	
	while(*tmp!='\\' && *tmp!='/' && *tmp!=0) ++tmp;
	if(start==tmp && start==name && *tmp!=0)
		curdir=sat3_home();		// IF NAME STARTS WITH A SLASH, BEGIN FROM HOME
	else {
		// CHECK IF IT'S UPDIR
		if( (tmp-start==2) && start[0]=='.' && start[1]=='.' && *tmp!=0) {
			// UPDIR	
			obj=__sat3_getdir(curdir);
			if(!obj) return 0;		// INVALID PATH
		} 
		else {
		
		obj=__sat3_rclindir(curdir,start,tmp-start);
		if(!obj) return 0;
		if(*tmp==0) return obj;				// RETURN OBJECT FOUND
		if(sat3_peek(obj,5)!=SAT_DORRP) return 0;
		}
		curdir=obj;
	}
		++tmp;
		start=tmp;
	}
return 0;	
}

// RCL ONLY IN CURRENT DIR

SAT_OBJECT sat3_rclhere(char *name)
{
return sat3_rclindir(sat3_getcwd(),name);
}


// RCL IN CURRENT DIRECTORY AND ALL DIRS ABOVE (LIKE CALCULATOR RCL)

SAT_OBJECT sat3_rcl(char *name)
{
	if(!name) return 0;
	
	int len;
	char *ptr=name;
	
	while(*ptr++);
	len=(int)(ptr-name);	
	
	SAT_OBJECT dir=sat3_getcwd(),obj;
	
	while(dir) {
	obj=sat3_rclindir(dir,name);
	if(obj) return obj;			// FOUND VARIABLE
	dir=__sat3_getdir(dir);
	}
	
	return 0;
	
}

// OBTAIN HIDDEN DIRECTORY POINTER

SAT_OBJECT sat3_hiddendir()
{
	return __sat3_rclindir(sat3_home(),NULL,0);
}


BOOL sat3_getpath(SAT_OBJECT obj,char *pathbuffer,int maxchars)
{
	int nchars;
	char *endptr;
	SAT_OBJECT curdir,home=sat3_home();
	
	if(maxchars<=0) return 0;
	*pathbuffer=0;
	endptr=pathbuffer+1;
	--maxchars;
	while(obj) {
		curdir=__sat3_getdir(obj);
		if(!curdir) return 0;
		if(curdir==home) {
			nchars=0;
			if(maxchars<nchars+1) return 0;
			memmove(pathbuffer+nchars+1,pathbuffer,endptr-pathbuffer);
			endptr+=nchars+1;
			pathbuffer[nchars]='/';	
			return 1;
		}
		else {
		nchars=sat3_peek(curdir-2,2);
		if(maxchars<nchars+1) return 0;
		memmove(pathbuffer+nchars+1,pathbuffer,endptr-pathbuffer);
		sat3_getvarname(curdir,pathbuffer,nchars+1);
		endptr+=nchars+1;
		pathbuffer[nchars]='/';	
		maxchars-=nchars+1;
		obj=curdir;
		}
	}
		
	return 0;
	}

	
#define SAT_SYSUPDSTART 0x806da	

// FIX USEROB POINTERS FOR ECUSER
// ALL PTRS >=startptr AND <endptr WILL BE UPDATED +offset

void __sat3_fixptrs2(SAT_OBJECT startptr,SAT_OBJECT endptr,SAT_OBJECT offset)
{
if((startptr&0x60000000)!=0) return;
if((endptr&0x60000000)!=0) return;

SAT_OBJECT tmpptr;
SAT_OBJECT chkptr;
int count;

// CHECK FOR POINTERS IN DATA STACK

tmpptr=sat3_peek(SAT_DSKTOP,5);
while((chkptr=sat3_peek(tmpptr,5))) {
if(chkptr>=startptr && chkptr<endptr) sat3_poke(tmpptr,chkptr+offset,5);	
tmpptr+=5;
}

// CHECK FOR POINTERS IN RETURN STACK

tmpptr=sat3_peek(SAT_RSKTOP,5)-5;
while( (chkptr=sat3_peek(tmpptr,5))& (~0xf)) {
if(chkptr>=startptr && chkptr<endptr) sat3_poke(tmpptr,chkptr+offset,5);	
tmpptr-=5;
}


// CHECK FOR TEMPORARY ENVIRONMENTS

tmpptr=sat3_peek(SAT_TEMPENV,5);


do {
	
count=sat3_peek(tmpptr,5);

tmpptr+=10;

if(count>15) {
	count-=15;
	
	do {
	chkptr=sat3_peek(tmpptr,5);
	if(chkptr>=startptr && chkptr<endptr) sat3_poke(tmpptr,chkptr+offset,5);	
	tmpptr+=5;
	chkptr=sat3_peek(tmpptr,5);
	if(chkptr>=startptr && chkptr<endptr) sat3_poke(tmpptr,chkptr+offset,5);	
	tmpptr+=5;
	count-=10;
	}
	while(count>0);
	
	count=1;	// ENSURE LOOP IS REPEATED
	continue; 
}
	
} while(count>0);


// CHECK FOR PTRS IN THE TOUCHTAB TABLE


tmpptr=sat3_peek(SAT_TOUCHTAB,5)+3;

while( (chkptr=sat3_peek(tmpptr,5))) {
if(chkptr>=startptr && chkptr<endptr) sat3_poke(tmpptr,chkptr+offset,5);	
tmpptr+=5;
}

// CHECK FOR PTRS IN THE FSTVGERPTR TABLE


tmpptr=sat3_peek(SAT_FSTVGERPTR,5);


if(tmpptr) {
	// CHECK FOR THE FSTVGERPTR ITSELF
	if(tmpptr>=startptr && tmpptr<endptr) sat3_poke(tmpptr,tmpptr+offset,5);	
	
	tmpptr+=15;
	count=sat3_peek(tmpptr,5);
	while(count--) {
		tmpptr+=20;
		while( (chkptr=sat3_peek(tmpptr,5))) {
			if(chkptr>=startptr && chkptr<endptr) sat3_poke(tmpptr,chkptr+offset,5);	
			tmpptr+=5;
		}
	}
	

}


// CHECK FOR SYSTEM VARIABLES

tmpptr=(SAT_OBJECT) SAT_SYSUPDSTART;

for(count=0x18B;count>=0;--count,tmpptr+=5)
{
chkptr=sat3_peek(tmpptr,5);
if(chkptr>=startptr && chkptr<endptr) sat3_poke(tmpptr,chkptr+offset,5);	
}

// ALL REFERENCES UPDATED

return;
}

// CHECK IF USEROB OBJECT IS REFERENCED

int __sat3_ptrisref2(SAT_OBJECT blk,SAT_OBJECT blkend)
{
if((blk&0x60000000)!=0) return INVALID_PTR;

SAT_OBJECT tmpptr;
SAT_OBJECT chkptr;
int count;

// CHECK FOR POINTERS IN DATA STACK

tmpptr=sat3_peek(SAT_DSKTOP,5);
while((chkptr=sat3_peek(tmpptr,5))) {
if(chkptr>=blk && chkptr<blkend) {
//	printf("%05X used in stack\n",chkptr);
	return BLOCK_REF;
}	
tmpptr+=5;
}

// CHECK FOR POINTERS IN RETURN STACK

tmpptr=sat3_peek(SAT_RSKTOP,5)-5;
while( (chkptr=sat3_peek(tmpptr,5))& (~0xf)) {
if(chkptr>=blk && chkptr<blkend) {
//	printf("%05X used in RSTK\n",chkptr);
	
	return BLOCK_REF;
}	
tmpptr-=5;
}

// CHECK FOR SYSTEM VARIABLES

tmpptr=(SAT_OBJECT) SAT_SYSUPDSTART;

for(count=0x18b;count>=0;--count,tmpptr+=5)
{
chkptr=sat3_peek(tmpptr,5);
if(chkptr>=blk && chkptr<blkend) {
//	printf("%05X used in SYSVARS %05X\n",chkptr,tmpptr);
	return BLOCK_REF;
}
}

// CHECK FOR TEMPORARY ENVIRONMENTS

tmpptr=sat3_peek(SAT_TEMPENV,5);


do {
	
count=sat3_peek(tmpptr,5);

tmpptr+=10;

if(count>15) {
	count-=15;
	
	do {
	chkptr=sat3_peek(tmpptr,5);
	tmpptr+=5;
	if(chkptr>=blk && chkptr<blkend) 
	{
//		printf("%05X used in TEMPENV\n",chkptr);
		
	return BLOCK_REF;
	}
	chkptr=sat3_peek(tmpptr,5);
	tmpptr+=5;
	if(chkptr>=blk && chkptr<blkend) {
//		printf("%05X used in TEMPENV\n",chkptr);
		return BLOCK_REF;
	}
	count-=10;
	}
	while(count>0);
	
	count=1;	// ENSURE LOOP IS REPEATED
	continue; 
}
	
} while(count>0);


// CHECK FOR PTRS IN THE TOUCHTAB TABLE


tmpptr=sat3_peek(SAT_TOUCHTAB,5)+3;

while( (chkptr=sat3_peek(tmpptr,5))) {
if(chkptr>=blk && chkptr<blkend) {
//	printf("%05X used in TOUCHTAB\n",chkptr);
	
	return BLOCK_REF;
}	
tmpptr+=5;
}

// CHECK FOR PTRS IN THE FSTVGERPTR TABLE


tmpptr=sat3_peek(SAT_FSTVGERPTR,5);


if(tmpptr) {
	// CHECK FOR THE FSTVGERPTR ITSELF
	if(tmpptr>=blk && tmpptr<blkend) {
//		printf("%05X used in FSTVGERPTR\n",tmpptr);
		
		return BLOCK_REF;
	}	
	
	tmpptr+=15;
	count=sat3_peek(tmpptr,5);
	while(count--) {
		tmpptr+=20;
		while( (chkptr=sat3_peek(tmpptr,5))) {
			if(chkptr>=blk && chkptr<blkend) {
//				printf("%05X used in FSTVGERPTR\n",chkptr);
				
				return BLOCK_REF;
			}	
			tmpptr+=5;
		}
	}
	

}

// BLOCK IS NOT REFERENCED

return BLOCK_UNREF;
}



void __sat3_fixdir(SAT_OBJECT dir,SAT_OBJECT position,int nibbles)
{

if(sat3_peek(dir,5)!=SAT_DORRP) return;
SAT_OBJECT dirend=sat3_skipob(dir),objend;
int offset;

dir=__sat3_dirskiplibs(dir);
if(!dir) return;
if(position<=dir || position>=dirend) return;

offset=sat3_peek(dir,5);
objend=dirend;
// PATCH OFFSET TO FIRST OBJECT
if(position>dir && position<dir+offset) {
	sat3_poke(dir,offset+nibbles,5);
}
offset=-offset;
while(offset!=0) {
dir-=offset;
if(position>=dir && position<objend) {
// PATCH DIRECTORY CHAIN
	if(objend!=dirend) {	// NO NEED TO PATCH IF FIRST OBJECT
		sat3_poke(objend-5,offset+nibbles,5);
	}
// RECURSIVELY PATCH OBJECT WITHIN DIRECTORY
__sat3_fixdir(sat3_sknameup(dir),position,nibbles);
return;
}
objend=dir;
dir-=5;
offset=sat3_peek(dir,5);	
}
}




// EXPAND/CONTRACT USEROB BY nibbles
// OPEN A HOLE AT position
// FIX ALL DIRECTORY STRUCTURES
// UPDATE ALL SYSTEM VARIABLES

BOOL sat3_ecuser(SAT_OBJECT position,int nibbles)
{
	if(!nibbles) return TRUE;
	SAT_OBJECT old_dsktop=sat3_peek(SAT_DSKTOP,5);
	
	if(sat3_getfreeuserob()<nibbles) return FALSE;
	
	// FIX DIRECTORY STRUCTURE
	__sat3_fixdir(sat3_home(),position,nibbles);
	
	// PATCH ALL POINTERS
	__sat3_fixptrs2(old_dsktop,position,-nibbles);
	
	sat3_memmove(old_dsktop-nibbles,old_dsktop,position-old_dsktop);
	return TRUE;
}

BOOL sat3_purgeobj(SAT_OBJECT varobject)
{
if(!varobject) return FALSE;


SAT_OBJECT dir=__sat3_getdir(varobject);
int nibbles;
SAT_OBJECT position;

if(!dir) return FALSE;		// NOT IN USEROB OR INVALID OBJECT
if(sat3_peek(dir,5)!=SAT_DORRP) return FALSE; // TRYING TO PURGE AND EMBEDDED OBJECT

position=sat3_sknamedn(varobject);
nibbles=sat3_skipob(varobject)-position;
int objused=__sat3_ptrisref2(position,position+nibbles);

if(objused==INVALID_PTR) return FALSE;
if(objused==BLOCK_REF) {
			// CHECK IF TRYING TO PURGE CURRENT DIRECTORY
			SAT_OBJECT context=sat3_getcwd();
			if(context>=position && context<position+nibbles) sat3_setcwd(dir); // CHANGE CURRENT DIR IF IT WILL BE REMOVED
			// CREATE COPY IN TEMPOB	
			SAT_OBJECT intempob=sat3_copyobject(ALLOC_TEMPOB,varobject);
			if(!intempob) return FALSE; // OBJECT IN USE CAN'T BE DEREFERENCED
			__sat3_fixptrs2(position,position+nibbles,intempob-varobject);	// MAKE ALL POINTERS GO TO THE COPY
}



nibbles+=5;
if(sat3_findfirst(dir)==varobject) {
	
// PURGING FIRST OBJECT IN DIR
	position-=5;
	SAT_OBJECT obj2=position-sat3_peek(position,5)+nibbles;
	dir+=nibbles;
	if(!sat3_ecuser(position,-nibbles)) return FALSE;
	
	// FIX FIRST OBJECT POINTER
	
	dir=__sat3_dirskiplibs(dir);
	
	sat3_poke(dir,(obj2-dir>5)? obj2-dir:0,5);		// POINT TO NEW LOCATION OF SECOND OBJECT
	
	return TRUE;
}
// PURGING IN THE MIDDLE OF A DIRECTORY	
return sat3_ecuser(position,-nibbles);	
	
}

BOOL sat3_purge(char *varpath)
{
return sat3_purgeobj(sat3_rclindir(sat3_getcwd(),varpath));
}

BOOL sat3_purgeindir(SAT_OBJECT dir,char *varname)
{
return sat3_purgeobj(sat3_rclindir(dir,varname));	
}


// STORE A VARIABLE IN GIVEN DIRECTORY
// DOES NOT CHECK FOR VALID NAME (NULL-NAMED AND ILLEGAL VARIABLES ARE OK)


BOOL sat3_stoindir(SAT_OBJECT dir,SAT_OBJECT obj,char *name)
{

	if(!dir) return FALSE;
	if(!obj) return FALSE;
	if(!name) return FALSE;
	if(sat3_peek(dir,5)!=SAT_DORRP) return FALSE;
	
	SAT_OBJECT var=0;
	
	// PARSE NAME FOR EMBEDDED PATH
	
	char *start,*tmp;
	SAT_OBJECT curdir=dir,obj2;
	tmp=name;
	start=tmp;
	
	while(*tmp!=0) {
	
	while(*tmp!='\\' && *tmp!='/' && *tmp!=0) ++tmp;
	if(start==tmp && start==name && *tmp!=0)
		curdir=sat3_home();		// IF NAME STARTS WITH A SLASH, BEGIN FROM HOME
	else {
		// CHECK IF UPDIR
		if( (tmp-start==2) && start[0]=='.' && start[1]=='.' && *tmp!=0) {
			obj2=__sat3_getdir(curdir);
			if(!obj2) return FALSE;	// INVALID NAME
		}
		else {
	    obj2=__sat3_rclindir(curdir,start,tmp-start);
//	    printf("rcl %s = %05X\n",start,obj2);
		if(!obj2) {
		if(*tmp==0) {	// LAST PORTION OF THE NAME	
			name=start;
			dir=curdir;
			break;
		}
		return FALSE;	// INVALID NAME OTHERWISE
		}
		if(*tmp==0) {
//			printf("current object found\n");
			var=obj2;				// OBJECT FOUND WITH SAME NAME, OVERWRITE
			dir=curdir;
			name=start;
			break;
		}
		if(sat3_peek(obj2,5)!=SAT_DORRP) return FALSE;	// INVALID PATH
		}
		curdir=obj2;
	}
		++tmp;
		start=tmp;
	}
	
	int objsize=sat3_sizeof(obj);
	
	// CHECK FOR OBJECT VALIDITY
	if(objsize<=5) return FALSE;
	// CHECK IF VARIABLE ALREADY EXISTS

	
	if(var) {
		// REPLACE OBJECT WITHOUT CHANGING NAME
		int currentsize=sat3_sizeof(var);
		SAT_OBJECT position=sat3_skipob(var);
		int objused=__sat3_ptrisref2(var,position);
		if(objused==BLOCK_REF) {
//			printf("object used\n");
			// CREATE COPY IN TEMPOB	
			SAT_OBJECT intempob=sat3_copyobject(ALLOC_TEMPOB,var);
			if(!intempob) return FALSE; // OBJECT IN USE CAN'T BE DEREFERENCED
			__sat3_fixptrs2(var,position,intempob-var);	// MAKE ALL POINTERS GO TO THE COPY
		}
//		printf("reallocate %05X,%05X\n",position,objsize-currentsize);
		if(!sat3_ecuser(position,objsize-currentsize)) return FALSE;	// EXTEND OR SHRINK AS NEEDED
//		printf("ecuser ok\n");
		var-=objsize-currentsize;
		sat3_memmove(var,obj,objsize);	
		return TRUE;
	}
	
	// CREATE NEW VARIABLE IN DIR
	
	SAT_OBJECT position=sat3_skipob(dir);
	dir=__sat3_dirskiplibs(dir);
	if(!dir) return FALSE;

	SAT_OBJECT firstob=dir+sat3_peek(dir,5);
	firstob=position-firstob;
	if(firstob<=5) firstob=0;		// END-OF-CHAIN MARKER	
	int namelen=strlen(name),nibbles=namelen;
	if(namelen) ++nibbles;
	++nibbles;
	nibbles<<=1;
	nibbles+=objsize+5;
	
	if(!sat3_ecuser(position,nibbles)) return FALSE;
	
	dir-=nibbles;
	position-=nibbles;
	sat3_poke(position,firstob,5);		// WRITE OFFSET CHAIN
	position+=5;
	sat3_poke(dir,position-dir,5);		// WRITE OFFSET TO FIRST OBJECT
	sat3_poke(position,namelen,2);		// WRITE VARIABLE NAME
	position+=2;
	if(namelen) {
	while(*name) { sat3_poke(position,*name,2); name++; position+=2; }	
	sat3_poke(position,namelen,2);	
	position+=2;
	}
	sat3_memmove(position,obj,objsize);	// WRITE OBJECT
	return TRUE;
}

// STORE IN CURENT DIRECTORY
BOOL sat3_sto(SAT_OBJECT obj,char *varname)
{
return sat3_stoindir(sat3_getcwd(),obj,varname);	
}


BOOL sat3_crdir(char *name)
{
SAT_OBJECT newdir=sat3_alloc(ALLOC_HEAP,13);

if(!newdir) return FALSE;

sat3_poke(newdir,SAT_DORRP,5);
sat3_poke(newdir+5,0x7ff,8);

int result=sat3_stoindir(sat3_getcwd(),newdir,name);

sat3_free(newdir);

return result;
}	


BOOL sat3_pgdir(char *name)
{
	return sat3_purge(name);
}

BOOL sat3_chdir(char *dir)
{
return sat3_setcwd(sat3_rclhere(dir));	
}



